using System;
using System.IO;
using System.Collections;

namespace StateMachines
{
	/// <summary>
	/// The test harness
	/// </summary>
	class MainClass
	{
		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main(string[] args)
		{
			// Scan the project directory
			foreach (FileSystemInfo Info in new AllFiles(Directory.GetCurrentDirectory() + @"\..\.."))
				Console.WriteLine(Info.FullName);
			Console.ReadLine();
		}
	}


	/// <summary>
	/// The interface that state handler classes must support.
	/// </summary>
	public interface IStateHandler
	{
		/// <summary>
		/// Just like IEnumerable.MoveNext, basically, 
		/// with a IStateMachine param to communicate with the state mechine object.
		/// </summary>
		/// <param name="StateMachine">Set next state and/or yield control to 'outer process'</param>
		/// <returns>false when no more items</returns>
		bool MoveNext(IStateMachine StateMachine) ;
	}

	/// <summary>
	/// State handlers interface to the state machine.
	/// </summary>
	public interface IStateMachine
	{
		/// <summary>
		/// This state has a Current value to set; we should yield now, and come back to this state again.
		/// </summary>
		/// <param name="NewValue">The new Current value</param>
		/// <returns>true</returns>
		bool Yield(object NewValue);

		/// <summary>
		/// This state has a new Current value to set; we should yield now, and come back to a different state next time.
		/// </summary>
		/// <param name="NewValue">The new Current value</param>
		/// <param name="NextState">The state to generate the next Current value</param>
		/// <returns>true</returns>
		bool Yield(object NewValue, IStateHandler NextState);

		/// <summary>
		/// This state is setting a new state; we should return that state's MoveNext().
		/// </summary>
		/// <param name="NextState">Next CurrentState</param>
		/// <returns>NextState.MoveNext()</returns>
		bool Recurse(IStateHandler NextState) ;
	}


	public abstract class Enumerator: IEnumerator, IStateMachine
	{
		/// <summary>
		/// Descendant classes override this to initialize the state machine.
		/// </summary>
		/// <returns>A state handler class.</returns>
		protected abstract IStateHandler StartState() ;

		#region IEnumerator Members
		public void Reset()
		{
			CurrentState = /*abstract*/ StartState();
			CurrentValue = null;
		}

		public object Current
		{	
			get	{	return CurrentValue;	}	
		}

		public bool MoveNext()
		{
			return CurrentState.MoveNext(this);
		}
		#endregion IEnumerator Members

		#region Private state machinery
		private IStateHandler CurrentState;
		private object CurrentValue;

		/// <summary>
		/// This state has a Current value to set; we should yield now, and come back to this state again.
		/// </summary>
		/// <param name="NewValue">The new Current value</param>
		/// <returns></returns>
		bool IStateMachine.Yield(object NewValue)
		{
			CurrentValue = NewValue;
			return true;
		}
		/// <summary>
		/// This state has a new Current value to set; we should yield now, and come back to a different state next time.
		/// </summary>
		/// <param name="NewValue">The new Current value</param>
		/// <param name="NextState">The state to generate the next Current value</param>
		/// <returns></returns>
		bool IStateMachine.Yield(object NewValue, IStateHandler NextState)
		{
			CurrentValue = NewValue;
			CurrentState = NextState;
			return true;
		}
		/// <summary>
		/// This state is setting a new state; we should return that state's MoveNext().
		/// </summary>
		/// <param name="NextState">Next CurrentState</param>
		/// <returns></returns>
		bool IStateMachine.Recurse(IStateHandler NextState)
		{
			if (NextState == null)
				return false; // no next state, no next item
			CurrentState = NextState;
			return NextState.MoveNext(this); // recurse
		}
		#endregion Private state machinery
	}

	/// <summary>
	/// "foreach (FileSystemInfo Info in new AllFiles(DirectoryNameOrInfo))" 
	/// does recursive visit - . directory, .\ files, then .\ subdirectories.
	/// You can use the AllFiles object as a one-off, or 
	/// save it and create multiple recursive iterators.
	/// </summary>
	class AllFiles: IEnumerable
	{
		#region RootDirectory & constructors
		private DirectoryInfo RootDirectory;

		public AllFiles(string RootDirectory): this(new DirectoryInfo(RootDirectory))
		{}

		public AllFiles(DirectoryInfo RootDirectory)
		{
			this.RootDirectory = RootDirectory;
		}
		#endregion RootDirectory & constructors

		public IEnumerator GetEnumerator()
		{
			return new DirectoryEnumerator(RootDirectory);
		}


		/// <summary>
		/// This implements a state machine with a state stack. 
		/// Reset() 'pushes' ShowDirectory with a null 'return state'.
		/// ShowDirectory returns FileSystemInfo on ., and transitions to ShowFiles.
		/// ShowFiles returns FileSystemInfo on each file, and transitions to ShowSubdirectories.
		/// ShowSubdirectories 'pushes' ShowDirectory, and pops back to controlling
		/// ShowSubdirectories if it can, and exits if it can't.
		/// </summary>
		private class DirectoryEnumerator: Enumerator
		{
			#region constructor, RootDirectory field
			private DirectoryInfo RootDirectory;

			public DirectoryEnumerator(DirectoryInfo RootDirectory)
			{
				this.RootDirectory = RootDirectory;
				Reset(); // foreach doesn't Reset()
			}
			#endregion constructor, RootDirectory field

			#region StartState override
protected override IStateHandler StartState()
{
	return new ShowDirectory(null, RootDirectory);
}
			#endregion StartState override

			#region Private state machinery
			#region Abstract classes that the state classes inherit from
			private abstract class baseState
			{
				protected IStateHandler NextState;

				protected baseState(IStateHandler NextState)
				{
					this.NextState = NextState;
				}
			}

			private abstract class arrayState: baseState
			{
				protected IEnumerator ArrayEnumerator;

				protected arrayState(IStateHandler NextState, IEnumerable Array): 
					base(NextState)
				{
					ArrayEnumerator = Array.GetEnumerator();
				}
			}

			#endregion Abstract classes that the state classes inherit from

			#region The three state classes
			private class ShowDirectory: baseState, IStateHandler
			{
				private DirectoryInfo Directory;

				public ShowDirectory(IStateHandler NextState, DirectoryInfo Directory): 
					base(NextState)
				{
					this.Directory = Directory;
				}

				public bool MoveNext(IStateMachine StateMachine)
				{
					IStateHandler subdirectories = 
						new ShowSubdirectories(NextState, Directory.GetDirectories());
					IStateHandler files = 
						new ShowFiles(subdirectories, Directory.GetFiles());

					// set Current to this directory; next time, scan files
					return StateMachine.Yield(Directory, files);
				}
			}

			private class ShowFiles: arrayState, IStateHandler
			{
				public ShowFiles(IStateHandler NextState, IEnumerable Array): 
					base(NextState, Array) {}

				public bool MoveNext(IStateMachine StateMachine)
				{
					if (ArrayEnumerator.MoveNext())
						return StateMachine.Yield((FileInfo) ArrayEnumerator.Current);
					else
						return StateMachine.Recurse(NextState);
				}
			}

			private class ShowSubdirectories: arrayState, IStateHandler
			{
				public ShowSubdirectories(IStateHandler NextState, 
					IEnumerable Array): 
					base(NextState, Array) {}

				public bool MoveNext(IStateMachine StateMachine)
				{
					if (ArrayEnumerator.MoveNext())
						// at least one more subdirectory - recurse
						return StateMachine.Recurse(
							new ShowDirectory(this, (DirectoryInfo) ArrayEnumerator.Current));
					else
						// no (more) subdirectories - pop
						return StateMachine.Recurse(NextState);
				}
			}
			#endregion The three state classes
			#endregion Private state machinery
		}
	}
}
